@media only screen and (min-width: 76.25em) {
  .md-main__inner {
    max-width: none;
    padding-left: 2em;
    padding-left: 2em;
  }

  .md-sidebar--primary {
    left: 0;
  }

  .md-sidebar--secondary {
    right: 0;
    margin-left: 0;
    -webkit-transform: none;
    transform: none;
  }
}