def test_import():
    import medchem  # noqa: F401
    import medchem.utils  # noqa: F401
    import medchem.groups  # noqa: F401
    import medchem.catalogs  # noqa: F401
    import medchem.constraints  # noqa: F401
    import medchem.complexity  # noqa: F401
    import medchem.rules  # noqa: F401
